# -*- mode:perl; fill-column:80 -*-
use utf8;

use 5.016;
use warnings;
use strict;

use Storable qw/dclone/;
use Carp qw/croak/;

$| = 1;

our @codemap;
our %codemap;
do "./trianguish-codes.pl" or die;

# Read the input.
#
# An input line starts with 0 or 2 spaces, and is followed by a pattern of three
# non-space characters, then a space, then three non-space characters, then a
# space, etc..
#
# The coordinate system is as follows:
#
#    0 1 2 3 4 5
# 0 ┌┴┐ ┌┴┐ ┌┴┐
# 0   └┬┘ └┬┘ └┬┘
# 1   ┌┴┐ ┌┴┐ ┌┴┐
# 1 └┬┘ └┬┘ └┬┘
# 2 ┌┴┐ ┌┴┐ ┌┴┐
# 2   └┬┘ └┬┘ └┬┘
# 3   ┌┴┐ ┌┴┐ ┌┴┐
# 3 └┬┘ └┬┘ └┬┘
#
# where the first coordinate (y) is the vertical position, and the second
# coordinate (x) is the horizontal position.
#
# To simplify things, the input must be translated to this position (i.e. the
# first, fourth, fifth, etc. lines touch the margin; the second, third, sixth,
# etc. lines don't).

my $odd_x = 0;
my $up_pointing = 1;
my $y = 0;
my $x_max = 0;

my @program;

while(<>) {
    next if /^$/;

    /^(  )?(\S\S\S(?: \S\S\S)*)\s*$/ or die
        "Input line is not in the expected format";
    !$1 == !$odd_x or die
        "Input line is not in the expected position";

    my $shifted_line = $2;
    my $x = $odd_x ? -1 : -2;
    $program[$y][$x += 2] = $_ for split / /, $shifted_line;
    $x > $x_max and $x_max = $x;

    $odd_x ^= 1 if $up_pointing;
    $up_pointing = !$up_pointing;
    ++$y if $up_pointing;
}


# How to produce debug output.
sub visualise {
    my $aref = shift;
    my $numerical = shift;

    output_text_grid(\*STDERR, $x_max, $aref, $numerical);
}

# Find the start point, represented as `***`.
# Also replace `###` with `undef`.
my ($start_y, $start_x);
for my $y (0 .. $#program) {
    my $line = $program[$y];
    for my $x (0 .. $#$line) {
        if ($line->[$x] eq '***') {
            defined $start_y and
                die "Two `***`: ($start_y, $start_x), ($y, $x)";
            $start_y = $y;
            $start_x = $x;
        }
        $line->[$x] eq '###' and $line->[$x] = undef;
    }
}
defined $start_y or die "No `***` to construct from";

print STDERR "Program to construct:\n";
visualise(\@program);

# For each point we need to construct, find its distance from the start.
sub find_distances {
    my $return_distances_in = shift;
    my $return_distance_order_in = shift;
    my $valid_cells = shift;
    my $init_y = shift;
    my $init_x = shift;
    @$return_distance_order_in = ($init_y, $init_x);
    @$return_distances_in = ();
    $return_distances_in->[$init_y][$init_x] = 0;

    my $i = 0;
    while ($i < @$return_distance_order_in) {
        my $y = $return_distance_order_in->[$i++];
        my $x = $return_distance_order_in->[$i++];
        my $distance = $return_distances_in->[$y][$x];

        my $add_distance = sub {
            my $new_y = shift;
            my $new_x = shift;
            return if $new_x < 0 || $new_y < 0;
            return unless defined $valid_cells->[$new_y][$new_x];

            unless (defined $return_distances_in->[$new_y][$new_x]) {
                push @$return_distance_order_in, $new_y, $new_x;
                $return_distances_in->[$new_y][$new_x] = $distance + 1;
            }
        };

        $add_distance->($y, $x-1);
        $add_distance->($y, $x+1);
        if (($x + $y) % 2) {
            $add_distance->($y+1, $x);
        } else {
            $add_distance->($y-1, $x);
        }
    }
}
my @distances;
my @distance_order;
find_distances \@distances, \@distance_order, \@program, $start_y, $start_x;

print STDERR "\nDistances:\n";
visualise(\@distances, 1);

# The cost of building on a given op. This is always a power of 2.
my @cost;
my $i = 0;
while ($i < @distance_order) {
    my $y = $distance_order[$i++];
    my $x = $distance_order[$i++];
    $cost[$y][$x] = 1 << $distances[$y][$x];
}

# The elbow that gives the lowest cost. This will be "<0>" if it's cheapest to
# build from the starting point. It is filled in as the elbows are scanned,
# rather than immediately.
my @build_elbow;

# Elbow locations.
my %elbow_y;
my %elbow_x;
my %shoulder_y;
my %shoulder_x;
# Cached information about elbows.
my %elbow_cells;
my %elbow_distances;
my %elbow_distance_orders;

# A function that returns an op's neighbours in the correct order (which is
# (lower x, equal x, higher x)).
sub neighbours {
    my $aref = shift;
    my $y = shift;
    my $x = shift;

    my $array_get = sub {
        my $new_y = shift;
        my $new_x = shift;
        $new_y < 0 || $new_x < 0 and return (undef);
        $aref->[$new_y][$new_x]
    };
    if (($x + $y) % 2) {
        ($array_get->($y, $x-1), $array_get->($y+1, $x), $array_get->($y, $x+1))
    } else {
        ($array_get->($y, $x-1), $array_get->($y-1, $x), $array_get->($y, $x+1))
    }
}

# Find the direction (l/v/r) of the preceding cell from this one.
sub closer_cell {
    my $distref = shift;
    my $y = shift;
    my $x = shift;
    my $tiebreak = shift; # 0 = clockwise, 1 = anticlockwise

    my ($l, $v, $r) = neighbours $distref, $y, $x;
    my $d = $distref->[$y][$x];
    defined $l and $l >= $d and $l = undef;
    defined $v and $v >= $d and $v = undef;
    defined $r and $r >= $d and $r = undef;

    my $prec =
        (defined($l) ? 1 : 0) * 1 +
        (defined($v) ? 1 : 0) * 2 +
        (defined($r) ? 1 : 0) * 4;
    my $t = ($tiebreak + $y + $x) % 2;

    $prec == 0 and return (undef);
    $prec == 7 and return "l"; # arbitrary
    if    ($prec == 1 || ($prec == 5 && !$t) || ($prec == 3 && $t)) { "l" }
    elsif ($prec == 2 || ($prec == 3 && !$t) || ($prec == 6 && $t)) { "v" }
    elsif ($prec == 4 || ($prec == 6 && !$t) || ($prec == 5 && $t)) { "r" }
}

# Modifies $$_[0], $$_[1] based on the l/v/r direction in $_[2].
sub add_direction {
    defined $_[0] or croak "Undefined add_direction y";
    defined $_[1] or croak "Undefined add_direction x";
    if ($_[2] eq 'l') { $_[1]--; }
    elsif ($_[2] eq 'v') { $_[0] += ($_[0] + $_[1]) % 2 ? 1 : -1; }
    elsif ($_[2] eq 'r') { $_[1]++; }
    else { croak "Unknown direction " . $_[2] }
}

# Modifies the relevant cell of the given text code.
sub modify_cell {
    my $cell_ref = shift;
    my $new_code = shift;
    my $dir = shift;
    my $invert = shift;
    $$cell_ref //= '---';
    if (($dir eq 'l' && !$invert) || ($dir eq 'r' && $invert)) {
        substr $$cell_ref, 0, 1, $new_code;
    } elsif ($dir eq 'v') {
        substr $$cell_ref, 1, 1, $new_code;
    } elsif (($dir eq 'r' && !$invert) || ($dir eq 'l' && $invert)) {
        substr $$cell_ref, 2, 1, $new_code;
    } else { croak "Unknown direction " . $dir }
}

# purely for debugging purposes, this isn't used elsewhere
print STDERR "\nPath back to starting cell:\n";
my @path_back;
$i = 0;
while ($i < @distance_order) {
    my $y = $distance_order[$i++];
    my $x = $distance_order[$i++];
    my $cc0 = closer_cell \@distances, $y, $x, 0;
    my $cc1 = closer_cell \@distances, $y, $x, 1;
    my $code = '---';
    defined $cc0 and modify_cell \$code, '0', $cc0;
    defined $cc1 and modify_cell \$code, '1', $cc1;
    $path_back[$y][$x] = $code;
}
visualise(\@path_back);


# Find cells that could be used as elbows.
# These are cells that have exactly two neighbours at a lower distance.
# Elbows are analysed in order from closest to most distant.
my $elbow_name = 'A';
$i = 0;
while ($i < @distance_order) {
    my $y = $distance_order[$i++];
    my $x = $distance_order[$i++];
    $build_elbow[$y][$x] //= "<0>";

    my $distance = $distances[$y][$x];
    my ($l, $v, $r) = neighbours \@distances, $y, $x;
    defined $l && defined $v && defined $r or next;
    $l eq $v && $v eq $r and next; # pathological but possible

    my $elbow;
    # tiebreak: "i" is furthest clockwise, "b" furthest anticlockwise
    if ($l eq $distance-1 && $v eq $distance-1) {
        $elbow = ($x + $y) % 2 ? "ibO" : "biO";
    } elsif ($v eq $distance-1 && $r eq $distance-1) {
        $elbow = ($x + $y) % 2 ? "Oib" : "Obi";
    } elsif ($l eq $distance - 1 && $r eq $distance-1) {
        $elbow = ($x + $y) % 2 ? "bOi" : "iOb";
    }
    $elbow or next;

    print STDERR "\nFound potential elbow $elbow_name:\n";

    # Find which cells are used to reach the elbow and which cells are buildable
    # from the elbow.

    # The paths to the elbow, plus the elbow itself.
    my @path; $#path = $#program;
    # Cells reachable from the elbow (only definedness matters).
    my @reachable = @{dclone(\@program)};
    my $path_y = $y;
    my $path_x = $x;
    my $build_cost;
    while (my $dir = closer_cell(\@distances, $path_y, $path_x, 0)) {
        $build_cost += 1 << $distances[$path_y][$path_x];
        modify_cell \$path[$path_y][$path_x], 'i', $dir;
        add_direction($path_y, $path_x, $dir);
        modify_cell \$path[$path_y][$path_x], 'o', $dir, 1;
        $reachable[$path_y][$path_x] = undef;
    }
    $path_y = $y;
    $path_x = $x;
    while (my $dir = closer_cell(\@distances, $path_y, $path_x, 1)) {
        modify_cell \$path[$path_y][$path_x], 'i', $dir;
        add_direction($path_y, $path_x, $dir);
        if (!$reachable[$path_y][$path_x]) {
            modify_cell \$path[$path_y][$path_x], 'O', $dir, 1;
            last;
        }
        modify_cell \$path[$path_y][$path_x], 'o', $dir, 1;
        $reachable[$path_y][$path_x] = undef;
        $build_cost += 1 << $distances[$path_y][$path_x];
    }
    $path[$y][$x] = $elbow;
    visualise(\@path);

    # Discover which cells are worth building from this elbow.

    my @elbow_distance;
    my @elbow_distance_order;
    find_distances \@elbow_distance, \@elbow_distance_order,
        \@reachable, $y, $x;
    print STDERR "Distances from elbow:\n";
    visualise(\@elbow_distance, 1);
    my $cost = $cost[$path_y][$path_x];
    print STDERR "Cost of building this elbow is $build_cost\n";
    print STDERR "Cost multiplier of building from this elbow is $cost\n";

    # Are the cost savings worth it?
    my $j = 0;
    my $savings = 0;
    while ($j < @elbow_distance_order) {
        my $cy = $elbow_distance_order[$j++];
        my $cx = $elbow_distance_order[$j++];
        my $d = $elbow_distance[$cy][$cx] - 1;
        if ($d >= 0 && $cost << $d < $cost[$cy][$cx]) {
            $savings += $cost[$cy][$cx] - ($cost << $d);
        }
    }
    print STDERR "Cost savings from building this elbow are $savings\n";

    if ($savings < $build_cost) {
        print STDERR "This elbow is not worth it.\n";
        next;
    }

    $j = 0;
    while ($j < @elbow_distance_order) {
        my $cy = $elbow_distance_order[$j++];
        my $cx = $elbow_distance_order[$j++];
        my $d = $elbow_distance[$cy][$cx] - 1;
        if ($d >= 0 && $cost << $d < $cost[$cy][$cx]) {
            $cost[$cy][$cx] = $cost << $d;
            $build_elbow[$cy][$cx] = "<$elbow_name>";
        }
    }

    print STDERR "Costs to build are now:\n";
    visualise(\@cost, 1);

    $elbow_y{"<$elbow_name>"} = $y;
    $elbow_x{"<$elbow_name>"} = $x;
    $shoulder_y{"<$elbow_name>"} = $path_y;
    $shoulder_x{"<$elbow_name>"} = $path_x;
    $elbow_cells{"<$elbow_name>"} = \@path;
    $elbow_distances{"<$elbow_name>"} = \@elbow_distance;
    $elbow_distance_orders{"<$elbow_name>"} = \@elbow_distance_order;

    ++$elbow_name;
}

print STDERR "\nCells are built from elbows:\n";
visualise(\@build_elbow);
print STDERR "\n";

# Work out the elbow sequencing.
#
# If an elbow X clobbers a cell that is built by elbow Y, then elbow X must
# be built before elbow Y is.
my %elbow_builds_before;
my %ebb_count;

for my $elbow (keys %elbow_cells) {
    $ebb_count{$elbow} //= 0;
    for my $y (0..$#program) {
        for my $x (0..$x_max) {
            next unless $elbow_cells{$elbow}[$y][$x];
            my $be = $build_elbow[$y][$x];
            unless ($elbow_builds_before{$elbow}{$be}) {
                print STDERR "Elbow build order: $elbow before $be\n";
                $elbow_builds_before{$elbow}{$be} = 1;
                $ebb_count{$be} //= 0;
                $ebb_count{$be}++;
            }
        }
    }
}

# TODO: This is wrong because $elbow_builds_before was not transitively closed
my @sorted_elbows =
    sort { $ebb_count{$a} <=> $ebb_count{$b} } sort {$b cmp $a} keys %ebb_count;
print STDERR "Elbow build order is thus [@sorted_elbows]\n";


# The actual construction.
#
# This is built out of a couple of mutually recursive functions:
#
# - build: builds the given op on the given cell;
# - build_wire: builds a wire to the given cell, with the given tiebreak.
#
# They output "build «code»" and "send «code»" to standard output.  (These are
# building and sending from the origin: to build/send from somewhere else, you
# add code that links the origin to it, e.g. to send from a particular cell, you
# build a wire to that cell and then send along the wire you built.)

$elbow_distances{"<0>"} = \@distances;
$elbow_distance_orders{"<0>"} = \@distance_order;

sub build;
sub build_wire;

# Builds on the given cell.
sub build {
    my $code = shift;
    my $y = shift;
    my $x = shift;
    my $via_elbow = shift;
    my $tiebreak = shift;

    # If on a downward-pointing triangle, reverse the code.
    ($y + $x) % 2 and $code = reverse $code;
    my $numcode = $codemap{$code};

    # Find the cell we're building from.
    my $direction = closer_cell $elbow_distances{$via_elbow}, $y, $x, $tiebreak;
    add_direction $y, $x, $direction;

    # Base case: building adjacent to the origin can be done directly.
    if ($distances[$y][$x] == 0) {
        print STDERR "\nbuild $code\n";
        print $numcode, ",";
        return;
    }

    # Special case: to build adjacent to the elbow, build a wire on its
    # shoulder, then a wire to its shoulder, then send the build command along
    # the wire.
    if ($elbow_distances{$via_elbow}[$y][$x] == 0) {
        my $sy = $shoulder_y{$via_elbow};
        my $sx = $shoulder_x{$via_elbow};
        my $shoulder_code = $elbow_cells{$via_elbow}[$sy][$sx];
        $shoulder_code =~ y/Oo/o-/;

        build $shoulder_code, $sy, $sx, "<0>", $tiebreak;
        build_wire $sy, $sx, "<0>", $tiebreak;
        print STDERR "\nsend $code\n";
        print $numcode, ",";
        return;
    }

    # Inductive case: build a builder on the previous cell, build a wire to the
    # builder, then send the build command along the wire.
    #
    # The neutral side is "o"; the side facing us "I"; the side facing the wire
    # "b".
    my $builder_code = "ooo";
    my $code_direction =
        closer_cell $elbow_distances{$via_elbow}, $y, $x, $tiebreak;
    modify_cell \$builder_code, "b", $code_direction;
    modify_cell \$builder_code, "I", $direction, 1;

    build $builder_code, $y, $x, $via_elbow, $tiebreak;
    build_wire $y, $x, $via_elbow, $tiebreak, $code_direction;
    print STDERR "\nsend $code\n";
    print $numcode, ",";
}

# Builds a wire up to the given cell, without changing the cell itself.
sub build_wire {
    my $y = shift;
    my $x = shift;
    my $via_elbow = shift;
    my $tiebreak = shift;
    my $input_direction = shift; # optional

    unless ($input_direction) {
        $input_direction =
            closer_cell $elbow_distances{$via_elbow}, $y, $x, $tiebreak;
    }

    if ($distances[$y][$x] == 0) { croak "Building a wire on the origin"; }

    # Base case: If adjacent to the origin, no wire is needed.
    if ($distances[$y][$x] == 1) {
        return;
    }

    # Special case: to build a wire to a cell adjacent to the elbow, build a
    # wire on its shoulder, then a wire to its shoulder.
    #
    # This is a pretty similar case to the special case in `build`, but the
    # details are different (the wire on the shoulder is built in a different
    # orientation and nothing is sent immediately).
    if ($elbow_distances{$via_elbow}[$y][$x] == 1) {
        my $sy = $shoulder_y{$via_elbow};
        my $sx = $shoulder_x{$via_elbow};
        my $shoulder_code = $elbow_cells{$via_elbow}[$sy][$sx];
        $shoulder_code =~ y/O/-/;

        build $shoulder_code, $sy, $sx, "<0>", $tiebreak;
        build_wire $sy, $sx, "<0>", $tiebreak;
        return;
    }

    # Inductive case: build a wire on the previous cell and a wire to the
    # previous cell.
    add_direction $y, $x, $input_direction;
    my $wire_code = "---";
    my $code_direction =
        closer_cell $elbow_distances{$via_elbow}, $y, $x, $tiebreak;
    modify_cell \$wire_code, "i", $code_direction;
    modify_cell \$wire_code, "o", $input_direction, 1;

    build $wire_code, $y, $x, $via_elbow, $tiebreak;
    build_wire $y, $x, $via_elbow, $tiebreak, $code_direction;
}

print "[";

for my $elbow (@sorted_elbows) {
    if ($elbow ne '<0>') {
        # Build the elbow.
        #
        # This involves building an O-builder at the point of the elbow, then a
        # wire to both of its inputs, using different tiebreaks to send the
        # wires along different paths.
        my $ey = $elbow_y{$elbow};
        my $ex = $elbow_x{$elbow};
        my $elbow_code = $elbow_cells{$elbow}[$ey][$ex];

        print STDERR "\nBuilding elbow $elbow: point of elbow:";
        build $elbow_code, $ey, $ex, "<0>", 1;
        # to simplify the code, do a spurious send after each of these wires is
        # built; the build wire has to be built first, so that we can clear the
        # cell the elbow is building on, so that when the spurious send down the
        # data wire happens the stray data doesn't end up clobbering anything
        print STDERR "\nBuilding elbow $elbow: build wire:";
        build_wire $ey, $ex, "<0>", 1;
        print STDERR "\nsend ...\n";
        print "0,";
        print STDERR "\nBuilding elbow $elbow: data wire:";
        build_wire $ey, $ex, "<0>", 0;
        print STDERR "\nsend ...\n";
        print "0,";
    }

    # Build from the elbow, in reverse distance order.
    print STDERR "\nBuilding from elbow $elbow:\n";
    while (@{$elbow_distance_orders{$elbow}}) {
        # note: popped in the reverse of the usual order
        my $x = pop @{$elbow_distance_orders{$elbow}};
        my $y = pop @{$elbow_distance_orders{$elbow}};
        if ($build_elbow[$y][$x] eq $elbow) {
            next if $program[$y][$x] eq '***';
            build $program[$y][$x], $y, $x, $elbow, 0;
        }
    }
}

# to make this loop properly, we need 9 36s as padding
print "36,36,36,36,36,36,36,36,36]\n";
